.386
.MODEL FLAT
;{{{  NOTES
; 386 kernel for occam
; This kernel has been developed by M D Poole with permission from
; kernels developed by INMOS Limited for various hosts and in particular
; one produced for VAX/VMS published at the Occam Retargetting Workshop
; at the University of Kent at Canterbury in September 1992.

; Licensees of versions of the Occam compiler for PC may adapt this
; kernel at their own risk. However it is not guaranteed that interfaces
; between the kernel and compiled code will remain as they are now,
; and so it is strongly recommended that any projects involving changes
; to this kernel are made known to the author who is always willing to
; discuss such projects. Coding conventions to ensure correct working of
; the kernel with the OLINK linker and code compiled from occam are not
; all explicit in this source.

; This kernel code is intended to be linked at load time with code
; compiled from occam. This code needs to jump to the start of the
; occam, and to be called through the jump table occam_table when the
; occam needs run-time support.

; There may also be a need to return to occam after some interrupts,
; but these may not be handled here?

; So that occam may access real hardware locations as PLACED variables
; or PORTs, it would be a useful simplification if CS and DS addressed
; 32-bit code and data segments mapped onto the whole of address space.
; This means that this assembly code cannot make the usual assumption
; that offsets of code and static data are relative to bases that will
; be at known offsets from the start of the segment.
;
; BUT!
; This version has been adapted to use the GO32 DOS extender.
; GO32 maps the first 1Mbyte of real memory on to DS:0xE0000000
; The compiler does not yet know about this (24/1/94)
;
;}}}  
;{{{  types and constants
npbyte TYPEDEF NEAR PTR BYTE; for indirect jump addresses

TRUE      EQU 1
FALSE     EQU 0
CR        EQU 13

BPW       EQU 4
BITSPW    EQU (BPW*8)
SHBW      EQU 2        ; log2 (BPW)
LoopCount EQU 100
poll_interval EQU 20
MaxHard    EQU 12      ; only 0 (kbd) and 1*BPW (scrn) coded so far
STD_IN_CHAN   EQU 0
STD_OUT_CHAN  EQU BPW
KSSIZE     EQU 1024    ; bytes of kernel stack

; termination status codes

term_normal      EQU  0
term_divzero     EQU  1
term_overflow    EQU  1
term_stop        EQU  2
term_deadlock    EQU  3
term_notout      EQU  4
term_notin       EQU  5
term_notalt      EQU  6
term_rangecheck  EQU  7
term_equalcheck  EQU  8
term_seterr      EQU  9
term_othertrap   EQU 10
term_badprotocol EQU 11
term_screen_not_word EQU 14
term_kbd_not_word EQU 15
;}}}  
;{{{  DOS constants - standard file handles
StdOutHandle     EQU 1
;}}}  

Okernel SEGMENT USE32
PROTECTED = 1
;--------------------------------------------------------------------------
;{{{  ; INIT (entry from operating environment) DOS + GO32 version
;-------------------------------------------------------------------------

OCCAMWS EQU (-4*BPW)     ; location of main_ws rel to KINIT
OCCAMVS EQU (-3*BPW)     ; location of main_vs rel to KINIT
KINIT:

; It is assumed that this is the entry point in 32-bit protected mode
; Certain information needs to be passed in and the following convention
; using registers is suggested:

; CS   set by GO32 to be the "all-code" segment selector
; DS   set by GO32 to be the "all-data" segment selector
; SS   set by this code = DS for kernel stack
; ES   set by this code = DS for local block moves
; FS   undefined
; GS   undefined
;
; EAX  undefined
; CX   undefined
; DX   undefined
; EBX  undefined
; ESP  stack pointer current value (C main prog parameter list)
; BP   undefined
; ESI  undefined
; EDI  undefined

; EIP  set by the far call to be the next instruction:

;
;This code is based on empirical observation of how a program run by GO32
;starts up, with ESP pointing to a C main program parameter list
;consisting of (argc, argv, envp) with usual C meaning of these matters.
;
; Extension of interface from go32 is achieved by pushing an extra
; (negative) parameter, before coming here
;

;{{{  get access to main program argument list (GO32 dependent)
	 pop        EAX             ; mdp_magic or argc
	 or         EAX, EAX        ; if <0 then can Set Int Vec
	 jge        setargc
	 mov        BYTE PTR CanSetIvec, TRUE
	 ; other communication from environment to kernel could be added here
	 pop        argc            ; count of command line strings
	 jmp        setargv
setargc:
	 mov        BYTE PTR CanSetIvec, FALSE
	 mov        argc, EAX       ; was count, not magic number
setargv:
	 pop        argv            ; pointer to array of arg pointers
	 pop        envp            ; pointer to array of env pointers
;}}}  

	 mov        EBP, -KSSIZE          ; default kstack size
	 add        EBP, ESP              ; occam workspace pointer
	 mov        AX, DS                ; Occam data segment (=code)
	 mov        SS, AX                ; stack segment = data segment
	 mov        ES, AX                ; data segment for movsb destinations
;
;  This is how space is allocated for the code compiled from occam:
;  Addresses may be real or virtual according to how the whole system works
;
;                   ------------------------
;            ESP -> | Kernel stack         |    Highest address
;                   |                      |
;                   |                      |
;                   ------------------------
;                   | Vector space         |
;                   |                      |
;                   |                      |
;                   ------------------------
;                   | Main program parms   |
;            EBP -> | Occam workspace      |
;                   |                      |
;                   ------------------------
;                   ~                      ~
;                   | Free space           |
;                   |                      |
;                   ------------------------
;                   ~                      ~
;                   |                      |
;                   | Compiled code        |
;                   |                      |
;                   ------------------------
;                   | Kernel comm area     |
;            ESI -> |                      |
;                   | Kernel entry table   |
;                   ------------------------
;                   | Kernel code          |
;                   ------------------------
;                   | Host OS comm area    |
;                   ------------------------     Lowest address
;
;Move info on vectorspace etc from below kernel code to locations
;near occam_start
	 mov        EAX, DWORD PTR KINIT[OCCAMVS]  ; vectorspace size words
	 shl        EAX, SHBW                      ; vectorspace size bytes
	 mov        vecsize, EAX
	 sub        EBP, EAX                    ; top of ws = bottom of vs
	 mov        vecbase, EBP
	 mov        freebase, 50000000h         ; PLACEHOLDER - free space pointer
	 mov        freesize, 100000h           ; PLACEHOLDER - free space size

 ;  Here any interrupt routines which run in 32-bit mode should be
 ;  identified by putting appropriate gate descriptors into the IDT
 ;  MAJOR MISSING AREA!!
 ;
	 sti                              ; allow interrupts again
;{{{  set divide error and overflow trap address
	 test       BYTE PTR CanSetIvec, TRUE ; test environment allows this
	 jz         @F
; This sequence assumes that the Set Interrupt Vector DOS function is
; trapped by the DOS-extender or whatever, so that it is not necessary to
; preserve real DOS interrupt vectors for subsequent reinstallation.
	 mov        EDX,  OFFSET Overflow
	 mov        AL, 0                   ; divide error
	 mov        AH, 25h                 ; Set Interrupt Vector
	 int        21h
	 mov        AL, 4                   ; interrupt on overflow
	 mov        AH, 25h                 ; Set Interrupt Vector
	 int        21h
@@:
;
;}}}  
	 call       InitQs             ; initialise process queue pointers
 ; set up parameters for standard hosted procedure in occam workspace
 ; start by allocating top of workspace as vector space
	 sub        EBP, 6*BPW                   ; parameter frame size
	 mov        EAX, vecbase                 ; vectorspace ptr
	 mov        DWORD PTR [EBP+5*BPW], EAX   ; vectorspace ptr
	 mov        EAX, freesize                ; free space size (words)
	 mov        DWORD PTR [EBP+4*BPW], EAX   ; free space size (words)
	 mov        EAX, freebase                ; free space pointer
	 mov        DWORD PTR [EBP+3*BPW], EAX   ; free space pointer
	 mov        DWORD PTR [EBP+2*BPW], STD_OUT_CHAN
	 mov        DWORD PTR [EBP+BPW], STD_IN_CHAN
	 cld                                ; direction should never change!
	 lea        ESI, occam_table        ; kernel entry point table
	 call       occam_start             ; return here at normal end
			     ; or elsewhere, by kernel call or exception trap
	 mov        AX, 4C00h               ; EXIT (0)
	 int        21h
;}}}  
;{{{  ; Interface to DOS and H/W (may need changes for other OSs)
; This seems to work on DOS + GO32, but may not be adequately generalised
; WHAT ASSUMPTIONS, IF ANY, SHOULD WE MAKE ABOUT OS's EFFECT ON REGISTERS?
;*************************************************************************
;{{{  GetHard
;-------------------------------------------------------------------------
;    GetHard
;            ECX = byte count, EBX = channel address, EDI = data address
;            All input from a hard channel is set up via GetHard
;            Entered from T_in code with return to occam in OccErrIptr
;            Also entered from PollRoutine if Inputs can be descheduled
;----------------------------------------------------------------------386
GetHard:
	 cmp      EBX, STD_IN_CHAN        ; Is it std_in (keyboard)?
	 jne      NotKeyboard
Keyboard:                       ; check byte count = 4 needed here!!!
	 push     EDI                     ; avoid CORRUPTION BY go32
	 mov      AH, 8                   ; Read keyboard without echo
	 int      21h
	 movzx    EAX, AL                 ; extend char to dword
	 pop      EDI
	 mov      [EDI], EAX              ; One character to data address
	 lea      ESI, occam_table        ; kernel entry point table
	 jmp      OccErrIptr

NotKeyboard:
; Here any other hard input channels should be coded!
	 mov      OccErrQual, EBX         ; Not a valid input channel
	 mov      OccErrCode, term_notin
	 jmp      ErrorExit

;}}}  
;{{{  PutHard
;--------------------------------------------------------------------------
;   PutHard
;            ECX = byte count, EBX = channel address, ESI = data address
;            All output to a hard channel goes via PutHard.
;-----------------------------------------------------------------------386

PutHard: cmp      EBX, STD_OUT_CHAN       ; Standard screen channel ?
	 jne      NotScreen
	 cmp      ECX, BPW                ; only words supported at present
	 je       CharToScreen
	 mov      OccErrQual, ECX
	 mov      OccErrCode, term_screen_not_word
	 jmp      ErrorExit
CharToScreen:
	 mov      EDX, [ESI]              ; word including char
	 or       EDX, EDX
	 jge      @F
	 mov      DL, CR                  ; special value -1 = end.of.line
@@:
	 mov      CharBuff, DL            ; save the char
	 lea      EDX, CharBuff           ; load its address
	 mov      EBX, StdOutHandle       ; DOS file handle
	 mov      ECX, 1                  ; count
	 mov      AH, 40h                 ; output 1 char to stdout
	 int      21h
	 lea      ESI, occam_table        ; kernel entry point table
	 jmp      OccErrIptr              ; return to occam

NotScreen:
; here should be coded any other hard output channels
	 mov      OccErrQual, EBX         ; Not a valid output channel
	 mov      OccErrCode, term_notout
	 jmp      ErrorExit

;}}}  
;{{{  TestHardChan
;-------------------------------------------------------------------------
;   TestHardChan
;                   On entry: EBX = channel number
;                   On exit:  EAX = channel status
;                   Assumed by PollRoutine not to corrupt EBX
;                   See if a hard channel is ready
;                   Called by call from T_in, T_Enbc, T_Disc or PollRoutine
;                   Can jump to ErrorExit if trying to ALT on invalid channel
;----------------------------------------------------------------------386
TestHardChan:
	 cmp      EBX, STD_IN_CHAN        ; Is it std_in (keyboard)?
	 jne      TNotKeyboard
TestKeyboard:
	 push     EDI                     ; in case it gets clobbered!
	 mov      AH, 0Bh                 ; Test keyboard status
	 int      21h
	 pop      EDI
	 and      EAX, 1                   ; convert to BOOL
	 ret

TNotKeyboard:
; here should be coded any other ALTable input channels!
	 mov         OccErrQual, EBX         ; Cannot alt on this channel
	 mov         OccErrCode, term_notalt
	 pop         EAX                     ; throw away return address
	 jmp         ErrorExit
;}}}  
;{{{  ReadTimer
;-------------------------------------------------------------------------
;    READ TIMER          on exit: Clock = current time
;-------------------------------------------------------------------------
; read time from opeating system
; there needs to be a monotonically increasing clock which goes back to
; 0 after getting to its maximum value.  NOT YET IMPLEMENTED
; Occam uses bottom 32 bits of such a clock.
;
; This version uses the ordinary DOS time of day clock.
;-------------------------------------------------------------------------

ReadTimer:
	 push        EDI
	 push        ESI
	 mov         AH, 2Ch                  ; DOS clock function
	 int         21h
	 movzx       EAX, CH                  ; hours
	 imul        EAX, 60
	 movzx       ECX, CL
	 add         EAX, ECX                 ; minutes
	 imul        EAX, 60
	 movzx       ECX, DH
	 add         EAX, ECX                 ; seconds
	 imul        EAX, 100
	 movzx       EDX, DL
	 add         EAX, EDX                 ; hundredths of seconds
	 imul        EAX, 10                  ; milliseconds
	 mov         Clock, EAX               ; time since midnight in ms
	 pop         ESI
	 pop         EDI
	 ret
;}}}  
;{{{  KickClock
;-------------------------------------------------------------------------
;  KICK CLOCK
;-------------------------------------------------------------------------
; Set up timer ast for the time in [ESP+2*BPW], this time must be
; some time in the future.
; Uses EAX(REG) as a temporary
;-------------------------------------------------------------------------

kickclock:
	 mov         AlarmSet, TRUE
	 mov         EAX, [ESP+2*BPW]
	 mov         AlarmTime, EAX
	 ret
;}}}  
;{{{  Clock "interrupt"
;
; routine entered when a desired time has elapsed and
; clock interrupts have been "enabled"
AST_CLOCK:
; entry here if operating system can asynchronously signal arrival
; of awaited alarm time.
;
; alarm clock timer.
; EDX is used to hold the value of the 32-bit OCCAM clock
; EBX is used to point at ws for process being considered
;
	 call        ReadTimer             ; time in Clock
	 mov         EDX, Clock            ; Now
TimeUp:
	 mov         AlarmSet, FALSE
	 jmp         DoTimeOut             ; in OS-independent code
;}}}  
;{{{  ShowHexDWord
; Diagnostic output of an 8 hex digit dword value to ErrOut
; On entry EBX contains DWord to be displayed as Hex digits
; also corrupts ECX, EDX, EDI
ShowHexDWord:
	   mov          ECX, 2*BPW                 ; digit count
	   xor          EDI, EDI                   ; clear EDI to index HexNumber
@@:
	   xor          EDX, EDX                   ; clear EDX
	   shld         EDX, EBX, 4                ; shift next digit to DL
	   shl          EBX, 4                     ; shift other digits also
	   mov          DL, HexDigits [EDX]        ; convert to ASCII Hex
	   mov          HexNumber [EDI], DL        ; store in digit sequence
	   inc          EDI                        ; increment the index
	   loopnz       @B                         ; loop decrementing ECX
	   mov          EBX, ErrOutHandle          ; DOS file handle
	   mov          ECX, 2*BPW+1               ; reset count incl a blank
	   lea          EDX, HexNumber             ; digit sequence
	   mov          AH, 40h                    ; write to file
	   int          21h
	   ret
;}}}  
;{{{  ShowProgName
; This code assumes the program name was passed as by GO32 and is pointed to
; by the pointer in argv.
; Having been used once its final character is changed to 'x' for dump file.
ShowProgName:
	mov       EAX, argv           ; pointer to pointer to progname
	mov       EDX, [EAX]          ; pointer to progname
	mov       ECX, 0              ; initialise count
@@:
	cmp       BYTE PTR [EDX + ECX], 0 ; is it a NUL?
	jz        @F
	inc       ECX                 ; increment count
	jmp       @B
@@:
	mov       EBX, ErrOutHandle   ; DOS file handle
	mov       AH, 40h             ; write file
	int       21h
	mov       BYTE PTR [EDX + ECX - 1], 'x' ; change name for dump file
	ret
;}}}  
;{{{  MemDump
; ------------------------------------------------------------------
; On entry:   BX = DOS File Handle
;             EDX = pointer to start of buffer
;             ECX = count of bytes to dump
; ------------------------------------------------------------------
MemDump:
	cmp       ECX, 4000h
	jle       DumpLastBlock
	push      ECX
	mov       ECX, 4000h
	mov       AH, 40h             ; write file
	int       21h
	pop       ECX
	sub       ECX, 4000h
	add       EDX, 4000h
	jmp       MemDump
DumpLastBlock:
	mov       AH, 40h             ; write file
	int       21h
	ret
;}}}  
;{{{  Hibernate
Hibernate:
; May need modification for 386 when we know more about interrupts!
; Purpose is to allow OS to schedule other tasks
; display of hibernate message may be useful in testing!
;        lea           EDX, Hiber_mess
;        mov           AH, 09h
;        int           21h                 ; display Hibernate! message
	jmp           SearchWork          ; Poll for input, clock, etc
;--------------------------------------------------------------------------
;}}}  
;{{{  Error Exit
;---------------------------------------------------------------------------
; ERROR exit from program
; STOP message to ErrOut
; short message to ERR file
; workspace dump to dump file
;---------------------------------------------------------------------------
Overflow:
	mov       OccErrIptr, EDI
	mov       OccErrCode, term_overflow
	jmp       ErrorExit
T_Seterr:
	pop       OccErrIptr
	mov       OccErrCode, term_seterr
	jmp       ErrorExit
RangeCheckError:
	pop       OccErrIptr
	mov       OccErrCode, term_rangecheck
ErrorExit:
	mov       EDX, OFFSET Errfilename
	xor       CX, CX                           ; ATTR_NORMAL
	mov       AH, 3Ch                          ; create file
	int       21h
	jc        FinalExit                        ; if fails to open
	mov       WORD PTR ErrOutHandle, AX        ; handle
	call      ShowProgName                     ; also marks last char 'x'
	movzx     EAX, BYTE PTR OccErrCode         ; ErrCode
	lea       EAX, HexDigits[EAX]              ; addr ErrCode as HexDigit
	mov       AL, [EAX]                        ; ErrCode as HexDigit
	mov       BYTE PTR (stopmsg + 8), AL       ; put in stopmsg
	mov       EBX, ErrOutHandle                ; DOS Handle
	mov       ECX, lstopmsg - 2                ; length of stopmsg
	lea       EDX, DWORD PTR (stopmsg + 2)
	mov       AH, 40h
	int       21h                              ; message to ErrOut
	mov       EBX, StdOutHandle                ; DOS Handle
	mov       ECX, lstopmsg - 1                ; length of stopmsg
	lea       EDX, DWORD PTR stopmsg
	mov       AH, 40h
	int       21h                              ; display message to StdOut
	mov       EBX, OccErrIptr                  ; occam iptr
	call      ShowHexDWord                     ; show as Hex number
	lea       EBX, occam_start                 ; occam start
	call      ShowHexDWord                     ; show as Hex number
	cmp       BYTE PTR OccErrCode, term_deadlock ; was it deadlock?
	jne       closeerrfile                     ; no, skip iotrace dump
iotracedump:
	mov       ECX, 16                          ; number of trace slots
	mov       EDI, iotptr                      ; most recently used
trlook:
	mov       EBX, [iotrace + 4*EDI]           ; channel ptr from trace
	or        EBX, EBX
	jz        trloop                           ; slot not used
	mov       EBX, [EBX]                       ; value of channel word
	cmp       EBX, NOT_PROCESS                 ; still waiting?
	jz        trloop
	mov       EBX, [EBX + W_Iptr]              ; Iptr at W[-1]
	push      EDI
	push      ECX
	call      ShowHexDWord                     ; show Iptr as hex number
	pop       ECX
	pop       EDI
	inc       BYTE PTR NumChansWaiting         ; count channels shown
	; note that duplicate occurrences of same Wptr should be ignored!
trloop:
	dec       EDI                              ; cyclically decrement EDI
	and       EDI, 0Fh
	loop      trlook
closeerrfile:                                      ; assumes file handle in EBX
	mov       AH, 3Eh                          ; close file
	int       21h
;{{{  create postmortem dump file
makedumpfile:
	mov       EAX, DWORD PTR KINIT[OCCAMWS]    ; workspace size words
	add       EAX, 6                           ; main prog params
	shl       EAX, SHBW                        ; workspace size bytes
	mov       ESI, vecbase                     ; bottom of vectorspace
	mov       EDX, ESI
	add       ESI, vecsize                     ; top of vectorspace
	sub       EDX, EAX                         ; workspace bottom
	sub       EDX, 5*BPW                       ; 5 below-workspace slots
	mov       ESP, EDX                         ; enable saving by pushes
; if we decide to save any more registers or kernel variables do so here!
	mov       AL, NumChansWaiting
	mov       dmptag + 3, AL                   ; put in byte 3 of dmptag
	cmp       BYTE PTR OccErrCode, term_deadlock ; was it deadlock?
	jne       afterchanloop                    ; no, skip iotrace dump
; save Wptrs for each waitingchannel, matching Iptrs saved above
	mov       ECX, 16                          ; number of trace slots
	mov       EDI, iotptr                      ; most recently used
trlookd:
	mov       EBX, [iotrace + 4*EDI]           ; channel ptr from trace
	or        EBX, EBX
	jz        trloopd                          ; slot not used
	mov       EBX, [EBX]                       ; value of channel word
	cmp       EBX, NOT_PROCESS                 ; still waiting?
	jz        trloopd
	push      EBX                              ; push Wptr of waiting io
trloopd:
	dec       EDI                              ; cyclically decrement EDI
	and       EDI, 0Fh
	loop      trlookd
afterchanloop:
; number of these pushes must match value in dmptag+2 (currently 14)
	push      DWORD PTR vecbase                ; save vecbase
	push      DWORD PTR vecsize                ; save vecsize
	push      EBP                              ; Wptr
	push      DWORD PTR OccErrIptr             ; Iptr at trap, etc
	lea       EAX, occam_start                 ; origin of Iptr values
	push      EAX
	push      DWORD PTR TimeQ                  ; queue ptrs
	push      DWORD PTR Back
	push      DWORD PTR Front
	push      DWORD PTR freebase               ; free base
	push      DWORD PTR freesize               ; free size (words)
	movzx     EAX, BYTE PTR OccErrCode         ; error code
	push      EAX
 ; finally the address and number of bytes of ws+vs dumped and the dmptag
	push      EDX                              ; address of first ws dumped
	mov       EAX, ESI
	sub       EAX, EDX
	push      EAX                              ; bytes dumped
	push      DWORD PTR dmptag                 ; identifier with counts
; get ready to write a dump of everything
	push      ESP                              ; bottom of dump
	push      ESI                              ; top of dump
	mov       EAX, argv                        ; pointer to pointer to progname
	mov       EDX, [EAX]                       ; pointer to (modified) progname
	xor       CX, CX                           ; ATTR_NORMAL
	mov       AH, 3Ch                          ; create file
	int       21h
	jc        FinalExit                        ; if fails to open
	mov       BX, AX                           ; file handle
	pop       ECX                              ; top of dump
	pop       EDX                              ; bottom of dump
	sub       ECX, EDX                         ; count of bytes to dump
	call      MemDump                          ; uses EDX, ECX, BX
	mov       AH, 3Eh                          ; close file
	int       21h
;}}}  
FinalExit:
	mov       AL, BYTE PTR OccErrCode
	mov       AH, 4Ch                          ; end program
	int       21h

; Needs thought for 386 traps etc
; ERROR HANDLER ; Catch divide by zero or explicit INTO as locatable error
		; All other traps give a message and exit ?!

;}}}  
;}}}  
;************************************************************************
;    All code below here is not dependent on DOS or any other particular
;    operating environment
;
;{{{  ; workspace constants

W_Iptr       EQU -BPW
W_Link       EQU W_Iptr-BPW
W_State      EQU W_Link-BPW
W_Tlink      EQU W_State-BPW
W_Time       EQU W_Tlink-BPW

; NB ** THE FOLLOWING FLAGS ASSUME VALUES 0 - 3 CANNOT BE ^code OR ^wspace
; (on T414 NOT_PROCESS is MinInt, but here it is 0)

NOT_SELECTED EQU 80000000h                    ; alt selection flag

NOT_PROCESS  EQU 0                           ; alt flags
Z_ENABLING   EQU NOT_PROCESS + 1
Z_WAITING    EQU NOT_PROCESS + 2
Z_READY      EQU NOT_PROCESS + 3
					     ; delay flags
Z_TIMESET    EQU NOT_PROCESS + 1
Z_TIMENOTSET EQU NOT_PROCESS + 2

;}}}  
;{{{  ; MACRO for QUEUE
;--------------------------------------------------------------------------
;   QUEUE            PROCESS (any register but EAX) is put on the queue
;-----------------------------------------------------------------------386
QUEUE    MACRO       PROCESS
	 LOCAL       L1,L2

	 cmp         Front, NOT_PROCESS    ; empty queue ?
	 jnz         L1
	 mov         Front, PROCESS        ; if only one add to front
	 jmp         L2                    ; and to the back
L1:      mov         EAX, Back
	 add         EAX, W_Link           ; queue is through link words
	 mov         [EAX], PROCESS        ; old back points to new
L2:      mov         Back, PROCESS         ; add process to back and return
ENDM
;}}}  
;{{{  ; Initialise queue pointers, looping counts, etc
InitQs:
	 mov        Front, NOT_PROCESS    ; initialise process queue
	 mov        TimeQ, NOT_PROCESS    ; initialise timer queue
	 mov        SoftStop, FALSE
	 mov        TimerInt, FALSE
	 mov        LoopWord, LoopCount
	 ret
;}}}  
;{{{  ; occam 2 Kernel
;************************************************************************
;------------------------------------------------------------------------
;{{{  ; comment on kernel register usage
;------------------------------------------------------------------------
;
; The kernel is called from occam via the kernel interface entry point table
; whose address is maintained in ESI in the compiled occam code
;              CALL occam_table [-BPW*entry]
;
; Wherever kernel code may corrupt ESI it is reloaded before return to occam.
; arguments (and results) are passed in [ESP] ....
;
; EBP is used by the occam code for its Workspace pointer
; ESP is used to address a small kernel stack, and should have been
;     returned to its entry value on exit.
; EBP, ESP are saved in Ostack, Kstack when necessary.
;
; All other 386 registers are free for kernel use when it is executing
;
;}}}  
;{{{  ; process queue comment
;-------------------------------------------------------------------------
; empty queue
;         Front = 0
;
; non-empty queue
;
;                   ;--------;    ;--------;    ;--------;
;       Front   --->|        | -->|        | -->|        |<---  Back
;                   ;--------; |  ;--------; |  ;--------;
;                   |        | |  |        | |  |        |
;                   ;--------; |  ;--------; |  ;--------;
;                   |  link  |--  |  link  |--  |  link  |
;                   ;--------;    ;--------;    ;--------;
;
;------------------------------------------------------------------------
; When any process is delayed or inputing from keyboard and there are no
; processes to run, STARTNEXTPROCESS should go to sleep
; until awakened by timer or keyboard.
;------------------------------------------------------------------------
;}}}  
;{{{  ; workspace layout comment
;--------------------------------------------------------------------------
; save area
;       ;--------;
; W --> |        | <---------- EMPTY or ^code (alt successor)
;       ;--------;
;       | iptr   | <---------- next process instruction
;       ;--------;
;       | link   | <---------- link pointer used in process queue
;       ;--------;
;       | State  | <---------- ENABLING or WAITING or READY or ^data
;       ;--------;
;       | tlink  | <---------- TIMESET or TIMENOTSET or ^wspace
;       ;--------;
;       | time   | <---------- process wake up time
;       ;--------;
;
; channel
;       ;--------;
;       |        | <---------- EMPTY or ^wspace (wdesc if priorities)
;       ;--------;
;--------------------------------------------------------------------------
;}}}  
;------------------------------------------------------------------------
;{{{  ; PollRoutine
;-------------------------------------------------------------------------

PollRoutine:
	    ; If hard channels can be made to interrupt, then the interrupt
	    ; routine should set a common flag which is tested here
	    ; to avoid unnecessary polling

	    ; Is there a process waiting for a hard input? (Kbd only at present)
	    test        BYTE PTR HardWaiting, TRUE
	    jz          @F
	    ; repeat following code for all waiting hard channels
;{{{  test keyboard and read char if one is waiting
	    mov         EBX, STD_IN_CHAN     ; Keyboard channel occam address
	    mov         EAX, [Chans0+EBX]    ; Should be Wptr of waiting process
	    cmp         EAX, NOT_PROCESS
	    je          @F                   ; should never go this way!
	    call        TestHardChan         ; poll (uses EBX, sets EAX)
	    cmp         EAX, FALSE           ; FALSE means not there yet
	    je          @F
	    dec         BYTE PTR HardWaiting ; clear waiting flag
	    mov         EBP, [Chans0+EBX]    ; Wptr of waiting process
	    mov         [Chans0+EBX], NOT_PROCESS  ; clear channel word
	    mov         ECX, BPW             ; Kbd behaves as CHAN OF INT
	    mov         EDI, [EBP+W_State]   ; saved data destination address
	    mov         EAX, [EBP+W_Iptr]    ; saved return address
	    mov         OccErrIptr, EAX      ; ready to return to occam
	    jmp         Keyboard
@@:
;}}}  
	    ; Is there a process waiting on the timer?
	    mov         EBX, TimeQ
	    cmp         EBX, NOT_PROCESS
	    jne         @F
	    ; Nobody waiting for anything, so don't bother to check time
	    ret
@@:
	    mov         ESI, [EBX+W_Time]    ; first time on queue
	    call        ReadTimer            ; read now
	    mov         EDX, Clock
	    cmp         ESI, EDX
	    jb          TimeUp              ; requires now in EDX
	    ret
;}}}  
;--------------------------------------------------------------------------
;}}}  
;{{{  ; Kernel calls from occam
;{{{  ;  RUNP    [ESP] ... = iptr, other process
;---------------------------------------------------------------------386
T_RunP:
	    pop         EDI
	    pop         EDX
	    jmp         queueit
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  STARTP  [ESP] ... = iptr, other process, relative iptr in that process
;---------------------------------------------------------------------386
T_StartP:
	    pop         EDI
	    pop         EDX                   ; Wdesc to be queued
	    pop         DWORD PTR [EDX+W_Iptr] ; relative Iptr of other process
	    add         [EDX+W_Iptr], EDI     ; put absolute Iptr in its Workspace
queueit:
	    QUEUE       EDX                   ; queue process
	    lea         ESI, occam_table      ; kernel entry point table
	    jmp         EDI                   ; continue current occam process
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  ENDP    [ESP] ... = iptr, successor workspace
;-------------------------------------------------------------------------
;
;   NB joincnt initially equal to processcount
;
;            ;--------;
;            |joincnt |<-- number of processes still to join
;            ;--------;
;     --->   |  iptr  |<-- set iptr to this when count becomes 0
;            |        |<-- set ^wspace here when count becomes 0
;            ;--------;
;
;---------------------------------------------------------------------386
T_EndP:
	    pop         EDI                ; continue address
	    pop         EBP                ; Wdesc for successor process
	    dec         DWORD PTR BPW[EBP]  ; decrement count (New W[1])
	    jnz         SearchWork         ; and test
					   ; run the continuation process
	    jmp         [EBP]              ; Starting at Iptr in W[0]
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  PAUSE   no parameters
;---------------------------------------------------------------------386
T_Pause:
	    dec         LoopWord
	    jle         @F
	    ret
@@:
	    pop         EDI
	    mov         DWORD PTR [EBP+W_Iptr], EDI  ; save iptr in workspace
	    QUEUE       EBP
SemiPoll:
	    dec         poll_count
	    jg          SearchWork1
SearchWork:                                ; start next process
	    push        EDI                ; return address for STOP diagnosis
	    call        PollRoutine
	    pop         EDI
	    mov         poll_count, poll_interval
SearchWork1:
	    cmp         Front, NOT_PROCESS ; any processes on the queue?
	    je          NothingNow         ; No
	    mov         EBP, Front         ; Yes - schedule it
	    cmp         EBP, Back          ; Anything else on the queue?
	    jne         @F                 ; Yes
	    mov         Front, NOT_PROCESS ; No, set queue empty
	    jmp         GoNew
@@:
	    mov         EBX, [EBP+W_link]  ; linked through link words
	    mov         Front, EBX         ; Point Front to next process

GoNew:
	    mov         LoopWord, LoopCount ; restart pause count
	    lea         ESI, occam_table    ; Kernel entry point table
	    jmp         [EBP+W_Iptr]        ; Enter the occam process found

NothingNow:                                 ; No processes to run
	    test        BYTE PTR HardWaiting, 1 ; Are we waiting for hard input?
	    jnz         Hibernate

	    cmp         TimeQ, NOT_PROCESS
	    jne         Hibernate          ; Process waiting on the timer

	    cmp         SoftStop, FALSE    ; Has a process stopped the kernel?
	    jne         @F                 ; Yes, (OccErrCode already set)

	    mov         OccErrCode,term_deadlock ; We have deadlock
	    mov         EDI, DeadLock      ; Make it visible
	    mov         OccErrIptr, EDI
@@:
	    jmp         ErrorExit
;---------------------------------------------------------------------------;
;}}}  
;{{{  ;  STOPP   no parameters
;----------------------------------------------------------------------386
T_StopP:
	    pop         EDI                ; occam address for STOP diagnosis
	    mov         DWORD PTR [EBP+W_Iptr],EDI; and do what Transputer does
	    cmp         SoftStop, FALSE
	    jnz         @F                ; earlier error detected

	    mov         SoftStop, TRUE    ; need to suppress this in RESCHEDULE !!!
	    mov         OccErrCode, term_stop
@@:
	    jmp         SearchWork
;---------------------------------------------------------------------------;
;}}}  
;{{{  ;  IN      [ESP] ... = iptr, byte count, channel address, data address     ;
;-----------------------------------------------------------------------386
T_In:
	    pop         OccErrIptr         ; save return iptr
	    pop         ECX                ; byte count
	    pop         EBX                ; channel address
	    pop         EDI                ; destination address
	    cmp         EBX, MaxHard       ; hard or soft channel?
	    ja          GetSoft            ; unsigned!
	    call        TestHardChan       ; is input waiting?
	    cmp         EAX, FALSE         ; FALSE means No!
	    jne         GetHard            ; get it and return to occam
	    lea         EBX, Chans0[EBX]   ; address to save BP
	    mov         BYTE PTR HardWaiting, TRUE; flag waiting condition
	    jmp         Waitin             ;
GetSoft:
	    mov         EDX, [EBX]         ; channel word
	    cmp         EDX, NOT_PROCESS   ; channel = EMPTY?
	    je          WaitIn
	    mov         ESI, [EDX+W_State] ; source address
;           jmp         ChanCopy           ; drop through to transfer data
;-------------------------------------------------------------------------
;                 ECX = byte count,
;                 EBX = channel address,
;                 ESI = source data address,
;                 EDI = destination address,
;                 EDX = Wdesc of other process
;---------------------------------------------------------------------386
ChanCopy:
	    cld
	    rep         movsb
	    mov         DWORD PTR [EBX], NOT_PROCESS   ; reset channel
	    mov         EDI, OccErrIptr    ; ready to return to occam
	    jmp         queueit            ; queue the other process (EDX)
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  OUTBYTE [ESP] ... = iptr, byte value, channel address                   ;
;---------------------------------------------------------------------386
T_OutByte:
	    pop         OccErrIptr         ; save return iptr
	    pop         [EBP]              ; byte value at W[0]
	    lea         ESI, [EBP]         ; pointer to it
	    pop         EBX                ; channel address
	    mov         ECX, 1             ; set byte count to one
	    jmp         Outcommon
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  OUTWORD [ESP] ... = iptr, word value, channel address                   ;
;---------------------------------------------------------------------386
T_OutWord:
	    pop         OccErrIptr         ; save return iptr
	    pop         DWORD PTR [EBP]    ; word value at W[0]
	    lea         ESI, [EBP]         ; pointer to it
	    pop         EBX                ; channel address
	    mov         ECX, BPW           ; set byte count to BPW
	    jmp         Outcommon
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  OUT     [ESP] ... = iptr, byte count, channel address, data address     ;
;---------------------------------------------------------------------386
T_Out:
	    pop         OccErrIptr         ; save return iptr
	    pop         ECX                ; byte count
	    pop         EBX                ; channel address
	    pop         ESI                ; source data address
Outcommon:
	    cmp         EBX, MaxHard       ; hard or soft channel?
	    jbe         PutHard            ; unsigned!
	    mov         EDX, [EBX]         ; channel word
	    cmp         EDX, NOT_PROCESS   ; channel = EMPTY?
	    jz          WaitOut
	    cmp         DWORD PTR [EDX+W_State], Z_READY ; other process in alt?
	    jbe         @F
	    mov         EDI, [EDX+W_State] ; destination address
	    jmp         ChanCopy           ; needs ECX, EBX, EDX, ESI, EDI
@@:
	    cmp         DWORD PTR [EDX+W_State], Z_WAITING  ; other process waiting?
	    jne         @F
	    mov         DWORD PTR [EDX+W_State], Z_READY
	    push        ESI                ; save source address
	    QUEUE       EDX                ; run other process
	    pop         ESI                ; recover source address
	    jmp         WaitOut            ; synchronise io
@@:
	    cmp         DWORD PTR [EDX+W_State], Z_ENABLING ; other process enabling?
	    jne         WaitOut
					   ; make other process alt ready
	    mov         DWORD PTR [EDX+W_State], Z_READY
WaitOut:
	    mov         [EBP+W_State], ESI  ; save source data address
	    jmp         WaitCommon
WaitIn:
	    mov         [EBP+W_State], EDI  ; save destination address
WaitCommon:
	    mov         EDI, OccErrIptr
	    mov         [EBP+W_Iptr], EDI   ; save return address
	    mov         [EBX], EBP          ; save Wdesc in channel
				      ; (kernel location for hard channel)
	    mov         EDI, iotptr
	    mov         [iotrace + 4*EDI], EBX  ; keep channel address in cyclic buffer
	    inc         EDI                 ; increment cyclic pointer
	    and         EDI, 0Fh            ; modulo 16
	    mov         iotptr, EDI
	    jmp         SearchWork
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  ENBC    [ESP] ... = iptr, guard, channel address, Creg
;          on exit: [ESP] ... = guard, Creg
;---------------------------------------------------------------------386
T_Enbc:
	    pop         OccErrIptr         ; Save return Iptr
	    pop         ECX                ; guard
	    pop         EBX                ; channel address
	    cmp         ECX, FALSE         ; test guard
	    jz          @F                 ; do nothing if FALSE
	    cmp         EBX, MaxHard       ; hard or soft channel?
	    ja          softchan           ; unsigned
; hardchan
	    cmp         DWORD PTR [EBP+W_State], Z_READY
	    je          @F
	    call        TestHardChan       ; poll (uses EBX, sets EAX)
	    cmp         EAX, FALSE         ; status returned in EAX
	    jz          @F
	    mov         DWORD PTR [EBP+W_State], Z_READY ; set ready
@@:
	    push        ECX                ; guard
	    jmp         OccErrIptr         ; continue with this process
softchan:
	    mov         EAX, [EBX]         ; channel word
	    cmp         EAX, NOT_PROCESS   ; channel = EMPTY?
	    jnz         @F
	    mov         [EBX], EBP         ; channel := W
	    jmp         continue           ; continue with this process
@@:
	    cmp         EAX, EBP           ; channel = W?
	    je          continue           ; allows for enables on same channel
					   ; (prevents cloning)
	    mov         DWORD PTR [EBP+W_State], Z_READY ; set ready
continue:
	    push        ECX                ; guard
	    lea         ESI, occam_table   ; Kernel entry point table
	    jmp         OccErrIptr         ; continue with this process
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  ALTWT   no parameters
;---------------------------------------------------------------------386
T_AltWt:
	    mov         DWORD PTR [EBP], NOT_SELECTED     ; W[0] := EMPTY
	    cmp         DWORD PTR [EBP+W_State], Z_READY  ; W[State] = READY?
	    jne         @F
	    ret
@@:
	    mov         DWORD PTR [EBP+W_State], Z_WAITING; W[State] := WAITING
	    pop         EDI                              ; for STOP diagnosis
	    mov         DWORD PTR [EBP+W_Iptr], EDI       ; save address
	    jmp         SearchWork
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  DISS    [ESP] ... = iptr, process address, guard, Creg
;          on exit: [ESP] ... = selected flag, Creg
;-------------------------------------------------------------------------
;    the first successful disable has its codeptr placed in location W[0]
;--------------------------------------------------------------------386

T_DisS:
DisableSkip:
	    pop         EDI                ; Save return Iptr
	    pop         EDX                ; process address
	    pop         ECX                ; guard
	    cmp         ECX, FALSE         ; do nothing if false
	    jz          returnfalse
SelectIfFirst:                             ; uses EBP, EDX, EDI
	    cmp         DWORD PTR [EBP], NOT_SELECTED  ; W[0] = EMPTY?
	    je          @F                       ; a branch chooses this one
returnfalse:
	    push        FALSE              ;
	    lea         ESI, occam_table   ; kernel entry point table
	    jmp         EDI                ; return to occam
@@:
	    mov         [EBP], EDX         ; W[0] := codeptr
	    push        TRUE
	    lea         ESI, occam_table   ; kernel entry point table
	    jmp         EDI
;-------------------------------------------------------------------------
;}}}  
;{{{  ;  DISC    [ESP] ... = iptr, process address, guard, channel address
;----------------------------------------------------------------------386
T_DisC:
DisableChan:
	    pop         EDI                ; Save return Iptr
	    pop         EDX                ; process address
	    pop         ECX                ; guard
	    pop         EBX                ; channel address
	    cmp         ECX, FALSE         ; do nothing if guard false
	    jz          returnfalse
	    cmp         EBX, MaxHard       ; hard or soft channel?
	    ja          disablesoft        ; unsigned
; hardchan
	    call        TestHardChan       ; return selected? in EAX
	    cmp         EAX, FALSE
	    jne         SelectIfFirst      ; uses EBP, EDX, EDI
	    jmp         returnfalse        ; return if not ready

disablesoft:
	    mov         EAX, [EBX]         ; channel word
	    cmp         EAX, NOT_PROCESS   ; channel = EMPTY?
	    jz          returnfalse
	    cmp         EAX, EBP           ; ^channel = W?
	    jne         SelectIfFirst
	    mov         DWORD PTR [EBX], NOT_PROCESS   ; channel := empty
	    jmp         returnfalse
;---------------------------------------------------------------------------
;}}}  
;{{{  ; Kernel calls, etc, using the time queue
;{{{  picture of queue
;---------------------------------------------------------------------------
;                                                        timer delay queue
; empty queue          tptr = 0
; non-empty queue (entries in order of increasing time into the future)
;
;            ;--------;    ;--------;     ;--------;     ;--------;
;            |  tptr  |--->|        |  *->|        |  *->|        |
;            ;--------;    ;--------;  |  ;--------;  |  ;--------;
;                          |  iptr  |  |  |  iptr  |  |  |  iptr  |
;                          ;--------;  |  ;--------;  |  ;--------;
;                          |  link  |  |  |  link  |  |  |  link  |
;                          ;--------;  |  ;--------;  |  ;--------;
;                          | State  |  |  |  State |  |  |  State |
;                          ;--------;  |  ;--------;  |  ;--------;
;                          |  tlink |--*  |  tlink |--*  | EMPTY=0|
;                          ;--------;     ;--------;     ;--------;
;                          |  time  |     |  time  |     |  time  |
;                          ;--------;     ;--------;     ;--------;
;
;---------------------------------------------------------------------------
;---------------------------------------------------------------------------
;}}}  
;{{{  ;   DoTimeOut
	 ; This code is entered with the current Clock time in EDX
	 ; Loop removes all processes made ready by
	 ; alarm call from the timer queue
DoTimeOut:
	 mov         EBX, TimeQ            ; front of time queue
	 cmp         EBX, NOT_PROCESS
	 jz          nomorepossible        ; Branch if nothing left on queue

	 mov         EAX, [EBX+W_Time]     ; Get time of process on queue
	 cmp         EAX, EDX              ; compare with now
	 ja          @F                    ; Branch if after (unsigned >)

	 ; transfer process from timer queue to processor queue
	 mov        EAX, [EBX+W_Tlink]
	 mov        TimeQ, EAX             ; Dequeue timer process
	 mov        DWORD PTR [EBX+W_Tlink], Z_TIMESET ; Set flag in workspace
	 cmp        DWORD PTR [EBX+W_State], Z_READY   ; Is process already ready?
	 je         DoTimeOut              ; Yes, go and do next one

	 mov        DWORD PTR [EBX+W_State], Z_READY   ; Set process ready
	 mov        [EBX+W_Time], EDX      ; Record current time
	 QUEUE      EBX                    ; Queue this process
	 jmp        DoTimeOut
@@:
	 push       DWORD PTR [EBX+W_Time] ; Wakeup time
	 push       EDX                    ; Now
	 call       TimeAfter              ; Set to wakeup then
	 pop        ECX                    ; after? (ALWAYS FALSE)
	 pop        EAX                    ; then - throw away
nomorepossible:
	 ret
;}}}  
;{{{  ;   INSERT
;-----------------------------------------------------------------------386
; insert current process into queue according to waiting time
; USES: EBX, EDX, ESI
insert:
	    lea         ESI, TimeQ         ; head of time queue ^tptr
tloop:
	    mov         EDX, [ESI]         ; dereference pointer
	    cmp         EDX, NOT_PROCESS   ; subs = EMPTY?
	    je          @F                 ; back end of queue
					   ; W[TIME] AFTER subs[TIME]?
	    mov         EBX, [EBP+W_Time]  ; W[TIME]
	    cmp         EBX, [EDX+W_Time]  ; subs[TIME]
	    jbe         @F                 ; slot into queue here
	    lea         ESI, [EDX+W_Tlink] ; prev := ^subs[TLINK]
	    jmp         tloop              ; loop
@@:
	    mov         [ESI], EBP         ; point it at W
	    mov         [EBP+W_Tlink], EDX ; W[TLINK] := subs
	    ret
;-------------------------------------------------------------------------
;}}}  
;{{{  ;   TIMEAFTER
;-------------------------------------------------------------------------
;                 On entry: [ESP] ... = return addr, now, then
;                 On exit:  [ESP] ... = after, then
;    where after is TRUE if now is AFTER then, otherwise FALSE
;    if FALSE, then clock has been kicked to interrupt at then
;                 Uses: EAX
;---------------------------------------------------------------------386
TimeAfter:
	    mov         EAX, [ESP+BPW]      ; now
	    cmp         EAX, [ESP+2*BPW]    ; compare with then
	    jbe         @F                  ; unsigned comparison <=
	    mov         DWORD PTR [ESP+BPW],TRUE  ; stack TRUE in place of now
	    ret                             ; pop return address from stack
@@:
	    mov         DWORD PTR [ESP+BPW], FALSE ; status is FALSE
	    cmp         AlarmSet, FALSE     ;
	    je          kickclock           ; alarm not yet set
	    mov         EAX, AlarmTime      ; look at when it is set for
	    cmp         EAX, [ESP+2*BPW]    ; compare with then
	    jb          kickclock           ; change alarm time
	    ret                             ; pop return address from stack
;}}}  
;{{{  ;   LDTIMER  leaves time at [ESP]
;---------------------------------------------------------------------386

T_LdTimer:
	    pop         EDI              ; return iptr
	    call        ReadTimer        ; returns time in Clock
	    push        Clock
					 ; we assume ESI not corrupted
	    jmp         EDI
;-------------------------------------------------------------------------
;}}}  
;{{{  ;   TIN      [ESP] ... = iptr, wait time
;----------------------------------------------------------------------386
T_Tin:
	    pop         EDI                ; return iptr
	    call        ReadTimer          ; read timer
	    push        Clock
	    call        TimeAfter          ; returns status in [ESP]
	    pop         ECX                ; after?
	    pop         EAX                ; then
	    or          ECX, ECX           ; time exceeded?
	    je          @F                 ; no?, wait
					   ; we assume ESI not corrupted
	    jmp         EDI                ; yes?, return

@@:         mov         DWORD PTR [EBP+W_State], Z_WAITING ; W[State] := WAITING
	    inc         EAX                ; then + 1
	    mov         [EBP+W_Time], EAX  ; W[TIME] := occamtime
	    call        insert
	    mov         [EBP+W_Iptr], EDI  ; W[IPTR] := iptr
	    jmp         SearchWork
;-------------------------------------------------------------------------
;}}}  
;{{{  ;   ENBT     [ESP] ... = iptr, guard, wait time, Creg
;          on exit: [ESP] ... = guard, Creg
;---------------------------------------------------------------------386
T_EnbT:
EnableTimeAfter:
	    pop         EDI                ; Save return Iptr
	    pop         ECX                ; guard
	    pop         EDX                ; wait time
	    cmp         ECX, FALSE         ; test guard
	    jz          enbtexit           ; do nothing if FALSE
	    cmp         DWORD PTR [EBP+W_Tlink], Z_TIMENOTSET ; W[TLINK] = TIMENOTSET?
	    jne         @F

	    mov         DWORD PTR [EBP+W_Tlink], Z_TIMESET    ; W[TLINK] := TIMESET
	    jmp         settime
@@:
	    cmp         DWORD PTR [EBP+W_Tlink], Z_TIMESET    ; W[TLINK] = TIMESET?
	    jne         enbtexit
	    mov         EAX, [EBP+W_Time]
	    cmp         EAX, EDX
	    jbe         enbtexit
settime:
	    mov         [EBP+W_Time],EDX   ; W[TIME] := exp
enbtexit:
	    push        ECX                ; guard
					   ; we assume ESI not corrupted
	    jmp         EDI                ; continue with this process
;------------------------------------------------------------------------
;}}}  
;{{{  ;   TALTWT   no parameters
;---------------------------------------------------------------------386
T_TAltWt:
	    pop         EDI                ; occam return address
	    call        ReadTimer          ; read time
	    mov         ESI, Clock         ; (clockreg)
	    mov         DWORD PTR [EBP], NOT_SELECTED  ; W[0] := EMPTY
	    cmp         DWORD PTR [EBP+W_State],Z_READY ; W[State] = READY?
	    je          wtaready           ; ready now
					   ; only return iptr in stack
	    cmp         DWORD PTR [EBP+W_Tlink],Z_TIMESET ; W[TLINK] = TIMESET?
	    jne         @F                 ; go and wait
	    push        [EBP+W_Time]       ; then
	    push        ESI                ; now
	    call        TimeAfter          ; status returned in [ESP]
	    pop         ECX
	    pop         EAX
	    or          ECX, ECX           ; time exceeded?
	    jne         wtaready           ; yes
@@:                                        ; otherwise wait
	    mov         DWORD PTR [EBP+W_State],Z_WAITING ; W[State] := WAITING
	    cmp         DWORD PTR [EBP+W_Tlink],Z_TIMESET ; W[TLINK] = TIMESET?
	    jne         @F
	    inc         DWORD PTR [EBP+W_Time] ; insert (W[TIME] + 1)
	    call        insert
@@:
	    mov         DWORD PTR [EBP+W_Iptr], EDI   ; save return iptr
	    jmp         SearchWork
wtaready:
	    mov         [EBP+W_Time], ESI  ; W[TIME] := clockreg
	    mov         DWORD PTR [EBP+W_State],Z_READY ; ready to go
	    lea         ESI, occam_table   ; kernel entry point table
	    jmp         EDI                ; continue current occam process
;-------------------------------------------------------------------------
;}}}  
;{{{  ;   DIST     [ESP] ... = iptr, process address, guard, wait time
;------------------------------------------------------------------------386
T_DisT:
DisableTimeAfter:
	    pop         EDI                ; Save return Iptr
	    pop         EDX                ; process address
	    pop         ECX                ; guard
	    pop         EAX                ; wait time
	    cmp         ECX, FALSE         ; do nothing if false
	    jz          returnfalse
	    cmp         DWORD PTR [EBP+W_Tlink], Z_TIMENOTSET    ; W[TLINK] = TIMENOTSET?
	    je          returnfalse        ; branch if a time set
	    cmp         DWORD PTR [EBP+W_Tlink], Z_TIMESET       ; W[TLINK] = TIMESET?
	    jne         remove             ; W[TLINK] = process
	    cmp         [EBP+W_Time], EAX  ; compare  W[TIME] and wait time
	    jle         returnfalse        ; j unless W[TIME] later
	    jmp         SelectIfFirst      ; uses EBP, EDX, EDI
remove:                                    ; uses ESI for prev, EBX for subs
	    lea         ESI,TimeQ          ; prev := ^tptr
distloop:
	    mov         EBX, [ESI]         ; subs := prev[0]
	    cmp         EBX, EBP           ; subs = W?
	    je          @F
	    lea         ESI, [EBX+W_Tlink] ; prev := ^subs[TLINK]
	    jmp         distloop           ; loop
@@:
	    mov         EBX, [EBP+W_Tlink] ; subs := W[TLINK]
	    mov         [ESI], EBX         ; prev[0] := subs
	    mov         DWORD PTR [EBP+W_Tlink], Z_TIMENOTSET   ; W[TLINK] = TIMENOTSET?
	    lea         ESI, occam_table
	    jmp         returnfalse       ; continue with this process
;----------------------------------------------------------------------------
;}}}  
;}}}  
;{{{  ; Floating point support  -- only Norm implemented!
; The target machine is flagged as not having the next 3 instructions
;
;{{{  Roundsn    [ESP] ... = iptr, guard word, fraction, exponent
;      on exit: [SP] = 32-bit REAL number
; Implementation of this instruction requires an understanding of IEEE 754!
T_Roundsn :
		ret
;}}}  
;{{{  Postnormsn [ESP] ... = iptr, guard word, high word, normalising shift
;                [BP] = exponent
; Implementation of this instruction requires an understanding of IEEE 754!
T_Postnormsn :
		ret
;}}}  
;{{{  Unpacksn   [ESP] ... = iptr, 32_bit real
; Implementation of this instruction requires an understanding of IEEE 754!
T_Unpacksn :
		ret
;}}}  
;{{{  Norm       [ESP] ... = iptr, lo, hi
;      On exit: [ESP] = lo, hi, count
; locates the highest 1 bit in [hi,lo], returns shifted double word and count
T_Norm :
		pop        EDI            ; return iptr
		pop        EAX            ; lo
		pop        EDX            ; hi
		mov        ECX, 31        ; 0 .. 31 in hi dword
		bsr        EBX, EDX       ; find highest bit set in hi
		jnz        shiftem
		mov        ECX, 63        ; 32 .. 63 in lo dword
		bsr        EBX, EAX       ; find highest bit set in lo
		jnz        longshift
		mov        ECX, 64        ; 2 whole dwords
		jmp        pushem
longshift:
		mov        EDX, EAX       ; lo into hi
		xor        EAX, EAX       ; clear lo
		sub        ECX, EBX       ; shift count + 32
		shl        EDX, CL        ; shift by (CL&#1F)
		jmp        pushem
shiftem:
		sub        ECX, EBX       ; generate shift count
		shld       EDX, EAX, CL   ; shift hi, filling from lo
		shl        EAX, CL        ; shift lo by same amount
pushem:
		push       ECX            ; count
		push       EDX            ; normalised hi
		push       EAX            ; normalised lo
		jmp        EDI            ; return to occam
;}}}  
;}}}  
;}}}  
;{{{  Data
;{{{  KernelData
; KERNEL DATA
;----------------------------------------------------------------------386
	    align dword
; kernel state variable slots
DpoolStart   dword       0A7BEADDEh   ; Magic number to aid disassembly
Back         dword       0
Front        dword       0
TimeQ        dword       0
LoopWord     dword       0
TimerInt     dword       0
SoftStop     dword       0
OccErrQual   dword       0
DeadLock     dword       0DEADDEADh   ; pseudo occam iptr for deadlocked programs
OccErrIptr   npbyte      0
ErrOutHandle dword       2
NumChansWaiting byte     0
OccErrCode   byte        term_normal   ; Error stuff is global
HardWaiting  byte        0          ; set when scheduler must check hard input
CanSetIvec   byte        0          ; passed in from calling environment
CharBuff     byte        0
HexDigits    byte        "0123456789ABCDEF"
HexNumber    byte        "         "
Errfilename  byte        "ERR",0
dmptag       byte        "dm", 14, 0
stopmsg      byte        13, 10, " STOP   @"
lstopmsg     equ         SIZEOF stopmsg
	     align       dword
Clock        dword       0
; hard channel vector
	    ; to hold W for hard channels (avoiding use of DUP!)
	    ; must be a word for any channel which might suspend on
	    ; attempting an In or an Out
Chans0       dword       0   ; Keyboard
Chans1       dword       0

poll_count   dword poll_interval
save_fp      dword  0
iotptr       dword   0
iotrace      dword  16 DUP (0)    ; used to note channels that may deadlock
;}}}  
;{{{  ; Timer Data
;-----------------------------------------------------------------------------
		align      dword
AlarmSet       dword       0
AlarmTime      dword       0
Hiber_mess     byte        "Hibernate! $"
;}}}  
;}}}  
;{{{  ; KERNEL jump table at end of data which comes after code
;-------------------------------------------------------------------------386

; the routines jumped to from here correspond closely (but not always exactly)
; to the corresponding T414 instructions
; kernel entry point table - addressed backwards from the end!
	 align       dword
; floating point support
; compiler knows which are not implemented and compiles library routine call
	 dword       T_Roundsn        ; 24      NOT IMPLEMENTED
	 dword       T_Postnormsn     ; 23      NOT IMPLEMENTED
	 dword       T_Unpacksn       ; 22      NOT IMPLEMENTED
	 dword       T_Norm           ; 21      84
; calls for error traps
	 dword       T_Seterr         ; 20      80
	 dword       RangeCheckError  ; 19      76
; alts
	 dword       T_DisT           ; 18      72
	 dword       T_DisC           ; 17      68
	 dword       T_DisS           ; 16      64
	 dword       T_EnbT           ; 15      60
	 dword       T_EnbC           ; 14      56
	 dword       T_TAltWt         ; 13      52
	 dword       T_AltWt          ; 12      48
	 dword       T_Tin            ; 11      44
	 dword       T_LdTimer        ; 10      40
; parallel
	 dword       T_Pause          ; 9       36
	 dword       T_StopP          ; 8       32
	 dword       T_EndP           ; 7       28
	 dword       T_StartP         ; 6       24
	 dword       T_RunP           ; 5       20
; io
	 dword       T_OutByte     ; 4          16
	 dword       T_OutWord     ; 3          12
	 dword       T_Out         ; 2           8
	 dword       T_In          ; 1           4
occam_table:
kcount   dword       23            ; number of entries

; The following routines do not exist as they are implemented by in-line code
; Note that this means the compiler and this code must have identical
; constants as special values in workspace special slots, etc
;         dword       Alt
;         dword       TAlt
;         dword       EnbS
;         dword       AltEnd
;}}}  
;{{{  ; extra communication area, for possible use by user ASM modules
; The following items are used to hold the base and size of vector space
; required by the program, and free space available to the program.
vecbase   dword       0
vecsize   dword       0
freebase  dword       0
freesize  dword       0

; The following items are used in a C supporting environment such as the go32
; run-time system in the DJGPP GNU C system.
; They enable user code to get access to environment, command line, etc
; through the occam_table pointer ESI
environment:
argc      dword       0
argv      dword       0
envp      dword       0
;}}}  
occam_start:
; the compiled code will be loaded to start here
Okernel ENDS
	 END      KINIT            ; entry point
